#!/usr/bin/env python

"""
Unit tests for the doorstop.vcs.base module.
"""

import unittest


class TestModule(unittest.TestCase):  # pylint: disable=R0904
    """Tests for the doorstop.vcs.base module."""  # pylint: disable=C0103

    def test_tbd(self):
        """Verify TBD."""
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()
